"""
NoORM (Not only ORM)

Typical ways to import:
- `import noorm.sqlite3 as nm` - for SQLite
- `import noorm.aiosqlite as nm` - for asynchronous SQLite via aiosqlite
- `import noorm.psycopg2 as nm` - for synchronous Postgres via psycopg2
- `import noorm.asyncpg as nm` - for asynchronous Postgres via asyncpg
- `import noorm.pymysql as nm` - for synchronous MySQL/MariaDB via PyMySQL
- `import noorm.aiomysql as nm` - for asynchronous MySQL/MariaDB via aiomysql
- `import noorm.sqlalchemy_sync as nm` - for synchronous SqlAlchemy
- `import noorm.sqlalchemy_async as nm` - for asynchronous SqlAlchemy
"""
